import json
import boto3
from datetime import datetime
import time
from lambda_actor.aws.sqs import *
import logging
from lambda_actor.types.type_conf import ActorConf
from lambda_actor.types.type_actor_message import *
from lambda_actor.utils.dateutil import timestamp


logger = logging.getLogger()

class ActorDriverException(Exception):
    pass

def actor_starter(bucket: str, prefix: str, conf_filename: str):
    """[Send task message]

    Args:
        bucket (str): [description]
        prefix (str): [description]
        conf_filename (str): [description]
    """
    # init
    s3_client = boto3.client('s3')
    actor_conf = ActorConf.get_actor_conf(s3_client=s3_client, bucket=bucket, prefix=prefix, conf_filename=conf_filename)
    trigger_input_path = f"{actor_conf.trigger_file_prefix}/{actor_conf.trigger_file}"
    tmp_trigger_input_path = f"/tmp/{actor_conf.trigger_file}"

    # sqs init
    sqs = boto3.resource('sqs', region_name='ap-northeast-1')
    executor_task_q = sqs.get_queue_by_name(QueueName=actor_conf.executor_task_q)
    
    # get csv
    s3_client.download_file(actor_conf.trigger_file_bucket, trigger_input_path, tmp_trigger_input_path)
    with open(tmp_trigger_input_path) as f:
            #executor_message_list = f.read().split('\n')[1:] #Skip header
            executor_message_list = f.read().split('\n')
            logger.info(f"executor_message_list: {executor_message_list}")
            
            # add executor task message
            executor_task_message_list = ExecutorTaskMessage.create_message(executor_message_list)
            logger.info(f"executor_task_message_list: {executor_task_message_list}")
            send_executor_task_message(queue=executor_task_q, message_body_list=ExecutorTaskMessage.encode_list(executor_task_message_list), executor_key=actor_conf.executor_key)

def executor_init_start(executor_concurrency: int, executor_key: str, executor_trigger_q):
    logger.info(f"executor_init_start")
    logger.info(f"executor_concurrency: {executor_concurrency}")
    for executor_id in range(executor_concurrency):
        executor_start(executor_id=executor_id, executor_key=executor_key, executor_trigger_q=executor_trigger_q)
        time.sleep(1)

def executor_start(executor_id: int, executor_key: str, executor_trigger_q,  driver_trigger_message: DriverTriggerMessage = None):
    logger.info(f"executor_start")
    logger.info(f"executor_id: {executor_id}")
    if driver_trigger_message:
        m = ExecutorTriggerMessage(status=ExecutorTriggerStatusType.START, message=f"executor {executor_id} start", driver_trigger_timestamp=driver_trigger_message.driver_trigger_timestamp, executor_id=executor_id)
    else:
        m = ExecutorTriggerMessage(status=ExecutorTriggerStatusType.INIT_START, message=f"executor {executor_id} start", driver_trigger_timestamp=timestamp(), executor_id=executor_id)
    send_execute_message(queue=executor_trigger_q, message_body=m.encode(), executor_id=executor_id, executor_key=executor_key)

def actor_driver(bucket: str, prefix: str, conf_filename: str, finally_func=None, driver_trigger_message_str: str = None):
    """[summary]

    Args:
        bucket (str): [description]
        prefix (str): [description]
        conf_filename (str): [description]
    """
    # init
    s3_client = boto3.client('s3')
    actor_conf = ActorConf.get_actor_conf(s3_client=s3_client, bucket=bucket, prefix=prefix, conf_filename=conf_filename)
    logger.info(f"actor_driver: actor driver start, {actor_conf}")

    # sqs init
    sqs = boto3.resource('sqs', region_name='ap-northeast-1')
    driver_trigger_q = sqs.get_queue_by_name(QueueName=actor_conf.driver_trigger_q)
    executor_trigger_q = sqs.get_queue_by_name(QueueName=actor_conf.executor_trigger_q)
    executor_task_q = sqs.get_queue_by_name(QueueName=actor_conf.executor_task_q)

    if driver_trigger_message_str is None:     # Init run
        actor_starter(bucket=bucket, prefix=prefix, conf_filename=conf_filename)
        executor_init_start(executor_concurrency=actor_conf.executor_concurrency, executor_key=actor_conf.executor_key, executor_trigger_q=executor_trigger_q)
    else:
        driver_trigger_message = DriverTriggerMessage.decode(driver_trigger_message_str)
        if (driver_trigger_message.status == DriverTriggerStatusType.CONTINUE) or (not is_q_empty(qname=actor_conf.executor_task_q)):
            print(f"contine: {driver_trigger_message}")
            executor_start(executor_id=driver_trigger_message.executor_id, executor_key=actor_conf.executor_key, executor_trigger_q=executor_trigger_q, driver_trigger_message=driver_trigger_message)
        elif driver_trigger_message.status == DriverTriggerStatusType.FINISH:
            print(f"actor_driver: DriverStatusType.FINISH, executor_id is {driver_trigger_message}")
            if finally_func:
                finally_func()
        else:
            raise ActorDriverException(f"driver_status_message.status  is illegal: {driver_status_message.status}")
    


    



from dataclasses import dataclass
from enum import Enum
from lambda_actor.utils.dateutil import timestamp
from typing import List
from dataclasses_json import dataclass_json
from datetime import datetime

class DriverTriggerStatusType(Enum):
    CONTINUE = "continue"
    FINISH = "finish"

@dataclass_json
@dataclass
class DriverTriggerMessage:
    """[summary]
    """
    status: DriverTriggerStatusType
    message: str
    executor_id: str
    driver_trigger_timestamp: str = timestamp()
    timestamp: str = datetime.now().strftime('%Y%m%d%H%M%S')

    @classmethod
    def decode(cls, message_str: str):
        """[summary]

        Args:
            message_str ([type]): [description]

        Returns:
            JobStatusMessage: [description]
        """
        return cls.from_json(message_str)
        
    def encode(self) -> str:
        """[summary]

        Returns:
            str: [description]
        """
        return self.to_json(indent=4, ensure_ascii=False)
       
class ExecutorTriggerStatusType(Enum):
    START = "start"
    INIT_START = "init_start"

@dataclass_json
@dataclass
class ExecutorTriggerMessage:
    """[summary]
    """
    status: ExecutorTriggerStatusType
    message: str
    driver_trigger_timestamp: str
    executor_id: int
    executor_trigger_timestamp: str = timestamp()
    timestamp: str = datetime.now().strftime('%Y%m%d%H%M%S')

    @classmethod
    def decode(cls, message_str: str):
        """[summary]

        Args:
            message_str ([type]): [description]

        Returns:
            JobStatusMessage: [description]
        """
        return cls.from_json(message_str)
        
    def encode(self) -> str:
        """[summary]

        Returns:
            str: [description]
        """
        return self.to_json(indent=4, ensure_ascii=False)
        #return f"{self.status.value},{self.message},{self.driver_trigger_timestamp},{self.executor_id},{self.executor_trigger_timestamp}"

@dataclass_json
@dataclass
class ExecutorTaskMessage:
    """[summary]
    """
    message: str
    retry_count: int = 0
    driver_start_timestamp: str = timestamp()
    timestamp: str = datetime.now().strftime('%Y%m%d%H%M%S')
    
    @classmethod
    def decode_list(cls, message_list: List[str]):
        """
        """
        return list(map(lambda m: ExecutorTaskMessage.decode(m), message_list))
        
    @classmethod
    def decode(cls, message_str: str):
        """[summary]

        Args:
            message_str (str): [description]

        Raises:
            Exception: [description]

        Returns:
            DriverType: [description]
        """
        return cls.from_json(message_str)
        
    def encode(self) -> str:
        """[summary]

        Returns:
            str: [description]
        """
        return self.to_json(indent=4, ensure_ascii=False)

    @classmethod
    def encode_list(cls, executor_task_message_list) -> str:
            """[summary]

            Returns:
                str: [description]
            """
            return list(map(lambda m: m.encode(), executor_task_message_list))

    @classmethod
    def create_message(cls, executor_message_list) -> str:
            """[summary]

            Returns:
                str: [description]
            """
            return list(map(lambda m: ExecutorTaskMessage(message=m), executor_message_list))


class ExecutorResultStatusType(Enum):
    SUCCESS = "success"
    FAILED = "failed"

@dataclass
class ExecutorResultMessage:
    """[summary]
    """
    status: ExecutorResultStatusType
    result: str
    driver_start_timestamp: str
    executor_start_timestamp: str
    retry_count: int
    executor_id: int
    execute_time: str
    timestamp: str = datetime.now().strftime('%Y%m%d%H%M%S')


from typing import List
import boto3

def send(queue, msg_list):
    size = 10
    for start in range(0, len(msg_list), size):
        queue.send_messages(Entries=msg_list[start: start + size])

def send_execute_message(queue, message_body: str, executor_id: int, executor_key: str):
    message_group_id = f"group_{executor_key}_{str(executor_id)}"
    msg_list = [{'Id': "0", 'MessageBody': message_body, 'MessageGroupId': message_group_id}]
    send(queue=queue, msg_list=msg_list)

def send_executor_task_message(queue, message_body_list: List[str], executor_key: str):
    msg_num = len(message_body_list)
    message_group_id = f"group_{executor_key}"
    msg_list = [{'Id': '{}'.format(i+1), 'MessageBody': message_body_list[i], 'MessageGroupId': message_group_id} for i in range(msg_num)]
    send(queue=queue, msg_list=msg_list)

def send_driver_trigger_message(queue, message_body: str, executor_id: int, executor_key: str):
    message_group_id = f"group_{executor_key}_{str(executor_id)}"
    msg_list = [{'Id': "0", 'MessageBody': message_body, 'MessageGroupId': message_group_id}]
    send(queue=queue, msg_list=msg_list)

def receive(queue, fetch_count=1) -> List[str]:
    msg_list = queue.receive_messages(MaxNumberOfMessages=fetch_count)
    def __get_q(m):
        s = m.body
        m.delete()
        return s
    message_body = list(map(lambda m: __get_q(m), msg_list))
    return message_body

def get(queue, fetch_count=1) -> List[str]:
    msg_list = queue.receive_messages(MaxNumberOfMessages=fetch_count)
    def __get_q(m):
        s = m.body
        return s
    message_body = list(map(lambda m: __get_q(m), msg_list))
    return message_body

def receive_all(queue) -> List[str]:
    message_body = []
    while True:
        # メッセージを取得
        msg_list = queue.receive_messages(MaxNumberOfMessages=10)
        if msg_list:
            for message in msg_list:
                message_body.append(message.body)
                message.delete()
        else:
            break
    return message_body

def get_q_current_size(qname: str):
    # ApproximateNumberOfMessages ApproximateNumberOfMessagesNotVisible ApproximateNumberOfMessagesDelayed
    m = boto3.client('sqs', region_name='ap-northeast-1').get_queue_attributes(QueueUrl=qname, AttributeNames=["ApproximateNumberOfMessages", "ApproximateNumberOfMessagesNotVisible", "ApproximateNumberOfMessagesDelayed"])
    message_count = m['Attributes']['ApproximateNumberOfMessages']
    return message_count

def is_q_empty(qname: str) -> bool:
    q_size = int(get_q_current_size(qname=qname))
    if q_size != 0:
        return True
    else:
        return False





import json
import boto3
from datetime import datetime
import time
from lambda_actor.aws.sqs import *
import logging
from lambda_actor.types.type_conf import ActorConf
from lambda_actor.types.type_actor_message import *

logger = logging.getLogger()

def clear_all_q(bucket: str, prefix: str, conf_filename: str):
    """
    """
    s3_client = boto3.client('s3')
    actor_conf = ActorConf.get_actor_conf(s3_client=s3_client, bucket=bucket, prefix=prefix, conf_filename=conf_filename)

    trigger_input_path = f"{actor_conf.trigger_file_prefix}/{actor_conf.trigger_file}"
    tmp_trigger_input_path = f"/tmp/{actor_conf.trigger_file}"
    logger.info(f"trigger_input_path: {trigger_input_path}")
    logger.info(f"tmp_trigger_input_path: {tmp_trigger_input_path}")
    logger.info(f"actor_conf: {actor_conf}")

    sqs = boto3.resource('sqs', region_name='ap-northeast-1')
    driver_trigger_q = sqs.get_queue_by_name(QueueName=actor_conf.driver_trigger_q)
    executor_trigger_q = sqs.get_queue_by_name(QueueName=actor_conf.executor_trigger_q)
    executor_task_q = sqs.get_queue_by_name(QueueName=actor_conf.executor_task_q)

    logger.info(f"clear driver_trigger_q: {receive_all(driver_trigger_q)}")
    logger.info(f"clear executor_trigger_q: {receive_all(executor_trigger_q)}")
    logger.info(f"clear executor_task_q: {receive_all(executor_task_q)}")


import json
import boto3
from datetime import datetime
import time
from lambda_actor.aws.sqs import *
import logging
from lambda_actor.types.type_conf import ActorConf
from lambda_actor.types.type_actor_message import *
from typing import List
import time

logger = logging.getLogger()


class ActorExecutorException(Exception):
    pass

def actor_executor(bucket: str, prefix: str, conf_filename: str, execution_func, finally_func = None, executor_trigger_message_str: str = None):
    """[summary]

    Args:
        bucket (str): [description]
        prefix (str): [description]
        conf_filename (str): [description]
    """
    # init
    s3_client = boto3.client('s3')
    actor_conf = ActorConf.get_actor_conf(s3_client=s3_client, bucket=bucket, prefix=prefix, conf_filename=conf_filename)
    logger.info(f"actor_conf: {actor_conf}")

    # sqs init
    sqs = boto3.resource('sqs', region_name='ap-northeast-1')
    driver_trigger_q = sqs.get_queue_by_name(QueueName=actor_conf.driver_trigger_q)
    executor_trigger_q = sqs.get_queue_by_name(QueueName=actor_conf.executor_trigger_q)
    executor_task_q = sqs.get_queue_by_name(QueueName=actor_conf.executor_task_q)

    # Run mannualy
    if executor_trigger_message_str is None:
        executor_trigger_message_str = receive(executor_trigger_q, 1)
        logger.info(executor_trigger_message_str)
        if len(executor_trigger_message_str) == 0:
            raise ActorExecutorException("executor_trigger_message_str is zero")
        executor_trigger_message = ExecutorTriggerMessage.decode(executor_trigger_message_str[0])
    else:
        executor_trigger_message = ExecutorTriggerMessage.decode(executor_trigger_message_str)

    # timer
    timeout = actor_conf.max_lambda_execution_time
    logger.info(f"executor timeout: {timeout}")
    executor_start = time.time()
    executor_result_list = []
    driver_trigger_status = DriverTriggerStatusType.CONTINUE

    while True:
        executer_task_message_str = receive(executor_task_q, 1)

        # Check task size
        if len(executer_task_message_str) == 0:
            logger.info(f"executor_task_q is zero, {executor_trigger_message}")
            driver_trigger_status = DriverTriggerStatusType.FINISH
            break

        executor_task_message = ExecutorTaskMessage.decode(executer_task_message_str[0])

        # run func
        task_start = time.time()
        executor_result_status = None
        result = None
        try:
            result = execution_func(executor_task_message.message)
            executor_result_status = ExecutorResultStatusType.SUCCESS
        except Exception as e:
            # when the task failed, retry
            retry_count = executor_task_message.retry_count + 1
            if retry_count > actor_conf.max_retry:
                result = str(e)
                executor_result_status = ExecutorResultStatusType.FAILED
                logger.error(f"actor_executor: {executor_task_message} is failed, {e}")
            else:
                executor_result_status = ExecutorResultStatusType.FAILED
                retry_executor_task_message = ExecutorTaskMessage(
                    message=executor_task_message.message,
                    retry_count=retry_count
                )
                retry_message = ExecutorTaskMessage.encode(retry_executor_task_message)
                send_executor_task_message(executor_task_q, [retry_message])
                logger.error(f"actor_executor: {retry_executor_task_message} is retried, {e}")

        # timer
        task_end = time.time()
        executed_time = task_end - executor_start
        task_time = task_end - task_start

        if result:
            # Create result
            executor_result = ExecutorResultMessage(
                status = executor_result_status,
                result = result,
                driver_start_timestamp = executor_task_message.driver_start_timestamp,
                executor_start_timestamp = executor_trigger_message.executor_trigger_timestamp,
                retry_count = executor_task_message.retry_count,
                executor_id = executor_trigger_message.executor_id,
                execute_time = task_time
            )
            executor_result_list.append(executor_result)
        if executed_time > timeout:
            logger.info(f"actor_executor is timeout, {executor_trigger_message}")
            break
        else:
            logger.debug(executed_time)

    # run own func
    if finally_func:
        finally_func(executor_result_list)

    # send finish message
    if driver_trigger_status == DriverTriggerStatusType.CONTINUE:
        retrun_driver_trigger_message = DriverTriggerMessage(
                status=DriverTriggerStatusType.CONTINUE,
                message=f"executor_id was {executor_trigger_message.executor_id}",
                executor_id=executor_trigger_message.executor_id
        )
    elif driver_trigger_status == DriverTriggerStatusType.FINISH:
        retrun_driver_trigger_message = DriverTriggerMessage(
                status=DriverTriggerStatusType.FINISH,
                message=f"executor_id was {executor_trigger_message.executor_id}",
                executor_id=executor_trigger_message.executor_id
        )
    else:
        raise ActorExecutorException("illegal driger status type")
    
    send_driver_trigger_message(queue=driver_trigger_q, message_body=retrun_driver_trigger_message.encode(), executor_id=executor_trigger_message.executor_id, executor_key=actor_conf.executor_key)

    


import logging
import sys
import os
import csv
import pandas as pd
from acdatalake.util.time import created_at
from datetime import datetime, timedelta, timezone
import json
from decimal import Decimal

class DecimalEncoder(json.JSONEncoder):
    def default(self, obj):
       if isinstance(obj, Decimal):
           return str(obj)
       return json.JSONEncoder.default(self, obj)

def write_df(s3_client, df: pd.DataFrame, bucket: str, path: str, filename: str, is_hist:bool = False) -> str:
    """[summary]

    Args:
        s3_client ([type]): [description]
        df (pd.DataFrame): [description]
        bucket (str): [description]
        path (str): [description]
        filename (str): [description]
        is_hist (bool, optional): [description]. Defaults to False.

    Returns:
        str: [description]
    """
    if path:
        fullpath = f'{path}/{filename}'
    else:
        fullpath = f'{filename}'
    df.to_csv(f'/tmp/{filename}.csv', sep=',', encoding='utf-8', index=False, quotechar='"', quoting=csv.QUOTE_ALL)
    if is_hist:
        fullpath = f"{fullpath}_{created_at()}.csv"
        s3_client.upload_file(f'/tmp/{filename}.csv', bucket, fullpath)
    else:
        fullpath = f'{fullpath}.csv'
        s3_client.upload_file(f'/tmp/{filename}.csv', bucket, fullpath)
    return f"s3://{bucket}/{fullpath}"
    
def write_dict(s3_client, d: dict, bucket: str, path: str, filename: str, _type: str="json", is_hist:bool = False) -> str:
    if path:
        fullpath = f'{path}/{filename}'
    else:
        fullpath = f'{filename}'
    with open(f'/tmp/{filename}.{_type}', 'w') as fp:
        json.dump(d, fp, ensure_ascii=False, cls=DecimalEncoder)
    if is_hist:
        fullpath = f"{fullpath}_{created_at()}.{_type}"
        s3_client.upload_file(f'/tmp/{filename}.{_type}', bucket, fullpath)
    else:
        fullpath = f'{fullpath}.{_type}'
        s3_client.upload_file(f'/tmp/{filename}.{_type}', bucket, fullpath)
    return f"s3://{bucket}/{fullpath}"
    
def read_df(s3_client, bucket: str, path: str, filename: str, created_at:datetime = None) -> pd.DataFrame:
    """[summary]

    Args:
        s3_client ([type]): [description]
        bucket (str): [description]
        path (str): [description]
        filename (str): [description]
        created_at (datetime, optional): [description]. Defaults to None.

    Returns:
        pd.DataFrame: [description]
    """
    if path:
        fullpath = f'{path}/{filename}'
    else:
        fullpath = f'{filename}'
    if created_at:
        s3_client.download_file(bucket, f'{fullpath}_{created_at.strftime("%Y%m%d")}.csv', f'/tmp/{filename}.csv')
        df = pd.read_csv(f'/tmp/{filename}.csv')
    else:
        s3_client.download_file(bucket, f'{fullpath}.csv', f'/tmp/{filename}.csv')
        df = pd.read_csv(f'/tmp/{filename}.csv')
    return df



from dataclasses import dataclass
import json


@dataclass
class ActorConf:
    """[summary]
    """
    driver_trigger_q: str
    executor_trigger_q: str
    executor_task_q: str
    executor_key: str
    executor_concurrency: int
    max_retry: int
    max_lambda_execution_time: int
    trigger_file_bucket: str
    trigger_file_prefix: str
    trigger_file: str

    @classmethod
    def get_actor_conf(cls, s3_client, bucket: str, prefix: str, conf_filename: str):
        """[summary]

        Args:
            s3_client ([type]): [description]
            bucket (str): [description]
            prefix (str): [description]
            conf_filename (str): [description]

        Returns:
            ActorConf: [description]
        """
        tmp_path = f'/tmp/{conf_filename}'
        s3_client.download_file(bucket, f'{prefix}/{conf_filename}', tmp_path)
        with open(tmp_path, encoding='utf-8') as f:
            actor_conf_json = json.load(f)
            actor_conf = ActorConf(
                driver_trigger_q=actor_conf_json["driver_trigger_q"],
                executor_trigger_q=actor_conf_json["executor_trigger_q"],
                executor_task_q=actor_conf_json["executor_task_q"],
                executor_key=actor_conf_json["executor_key"],
                executor_concurrency=actor_conf_json["executor_concurrency"],
                max_retry=actor_conf_json["max_retry"],
                max_lambda_execution_time=actor_conf_json["max_lambda_execution_time"],
                trigger_file_bucket=actor_conf_json["trigger_file_bucket"],
                trigger_file_prefix=actor_conf_json["trigger_file_prefix"],
                trigger_file = actor_conf_json["trigger_file"],
                )
        return actor_conf




from datetime import datetime, timedelta, timezone, date
from decimal import Decimal
import time
import calendar

def daterange(_start, _end):
    for n in range((_end - _start).days):
        yield _start + timedelta(n)

def created_at(is_month: bool =False):
    JST = timezone(timedelta(hours=+9), 'JST')
    timestamp = Decimal(time.time())
    today = datetime.now(JST).date()
    if is_month:
        created_at = str(today.strftime("%Y%m"))
    else:
        created_at = str(today.strftime("%Y%m%d"))
    return created_at

def timestamp():
    now = datetime.now(timezone(timedelta(hours=9))) # 日本時刻
    s = now.strftime('%Y/%m/%d %H:%M:%S')  # yyyyMMddHHmmss形式で出力
    return s
    
def today():
    return created_at()

def tomorrow():
    JST = timezone(timedelta(hours=+9), 'JST')
    today = datetime.now(JST).date()
    return str((today + timedelta(days=1)).strftime("%Y%m%d"))
    
def yesterday():
    JST = timezone(timedelta(hours=+9), 'JST')
    today = datetime.now(JST).date()
    return str((today - timedelta(days=1)).strftime("%Y%m%d"))
    

def times():
    JST = timezone(timedelta(hours=+9), 'JST')
    timestamp = Decimal(time.time())
    today = datetime.now(JST).date()
    weekday = today.strftime('%a')
    created_at = str(today.strftime("%Y%m%d"))
    created_at_unix = Decimal(time.mktime(today.timetuple()))
    last_created_at = str((today - timedelta(days=1)).strftime("%Y%m%d"))
    return {"timestamp": timestamp, "created_at": created_at, "created_at_unix": created_at_unix,\
            "weekday": weekday, "last_created_at": last_created_at}

def get_first_date2(year, month):
    return date(int(year), int(month), 1)
    
def get_last_date2(year, month):
    return date(int(year), int(month), calendar.monthrange(int(year), int(month))[1])